#!/bin/python

from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .value import MBIOValue

from .task import MBIOTask
from .xmlconfig import XMLConfig
from .xknxclient import XKNXClient


class MBIOTaskKnx(MBIOTask):
    def initName(self):
        return 'knx'

    @property
    def client(self):
        return self._client

    def onInit(self):
        self._client=None
        self._data={}

        self.config.set('refreshperiod', 10)
        self._timeoutRefresh=0
        self.valueDigital('comerr', default=False)

    def addGroup(self, name, address, dtype, timeout, writable=False):
        if name and name not in self._data and address:
            data={}
            data['address']=address
            data['knx']=self._client._values.add(address, timeout=timeout)
            if dtype in ('bool'):
                value=self.valueDigital('%s_state' % name, writable=writable, commissionable=True)
            elif dtype in ('byte', 'int', 'number', 'float'):
                value=self.value('%s_value' % name, writable=writable, commissionable=True)
            data['state']=value

            self.logger.debug("KNX: mapping knx %s group %s to value %s", dtype, address, value.name)
            self._data[name]=data
            return value

    def onLoad(self, xml: XMLConfig):
        self.config.update('refreshperiod', xml.getInt('refresh'))
        self.config.set('host', xml.get('host'))
        if not self.config.host:
            self.config.update('host', self.pickleRead('host'))

        self._client=XKNXClient(host=self.config.host,
                logger=self.logger,
                timeoutRefresh=self.config.refreshperiod)

        items=xml.children('group')
        if items:
            for item in items:
                # deny name duplication
                name=item.get('name')
                address=item.get('address')
                dtype=item.get('type')
                timeout=item.getFloat('timeout')
                writable=item.getBool('writable')
                self.addGroup(name, address, dtype, timeout, writable)

    def waitForConnection(self, timeout=2.0):
        if self.client and self.config.host:
            self.client.start()
            timeout=self.timeout(timeout)
            while not self.isTimeout(timeout):
                if self.client.isConnected():
                    self.logger.info('KNX: connected to router %s' % self.config.host)
                    return True
                self.sleep(0.1)

        self.logger.error('KNX: unable to connect to router %s' % self.config.host)
        return False

    def poweron(self):
        if self.waitForConnection(3.0):
            self.client.refreshValues()
        return True

    def poweroff(self):
        self.client.stop()
        return True

    def run(self):
        delay=5.0

        if not self.client.isConnected():
            self.values.comerr.updateValue(True)
            for data in self._data.values():
                value=data['state']
                value.setError(True)
        else:
            self.values.comerr.updateValue(False)
            for data in self._data.values():
                knxvalue=data['knx']
                knxvalue.refreshIfTimeout()

                if knxvalue.age()>180:
                    value.setError(True)
                else:
                    value=data['state']
                    value.updateValue(knxvalue.data)
                    value.setError(False)
                    if value.isPendingSync():
                        knxvalue.write(value.toReachValue)
                        value.clearSync()
                        delay=1.0

        return delay

    def discover(self, stopOnFound=False):
        return self.client.discover(self.getMBIO().interface, stopOnFound=stopOnFound)


if __name__ == "__main__":
    pass



#!/bin/python

from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .value import MBIOValue

from .task import MBIOTask
from .xmlconfig import XMLConfig


# During the onInit() phase, the task pre-declares inputs
# During the onLoad() phase, a ref (mbio-value, constant) is looked for each predeclared value
# A mbio-value is declared for each input which is still unlinked after the load phase

class VIO(object):
    def __init__(self):
        pass


class VBlock(object):
    def __init__(self):
        pass


class MBIOTaskVPLC(MBIOTask):
    def initName(self):
        name=self.__class__.__name__.lower().removeprefix('mbiotask')
        return '%s%d' % (name, self._parent.tasks.countByType(self.__class__))

    def declareAnalogInput(self, name, unit, resolution=0.1, default=None):
        if name:
            name=name.lower()
            if name not in self._inputs:
                i={'name': name, 'type': 'analog', 'value': None, 'unit': None, 'resolution': resolution, 'default': default}
                self._inputs[name]=i

    def declareDigitalInput(self, name, default=None):
        if name:
            name=name.lower()
            if name not in self._inputs:
                i={'name': name, 'type': 'digital', 'value': None, 'default': default}
                self._inputs[name]=i

    def declareInputs(self):
        # to be overriden
        pass

    def declareAnalogOutput(self, name, unit, resolution=0.1, default=None):
        if name:
            name=name.lower()
            if name not in self._outputs:
                self._outputs[name]=self.value('o_%s' % name, unit=unit, resolution=resolution, default=default)

    def declareDigitalOutput(self, name, default=None):
        if name:
            name=name.lower()
            if name not in self._outputs:
                self._outputs[name]=self.valueDigital('o_%s' % name, default=default)

    def declareOutputs(self):
        # to be overriden
        pass

    def onInit(self):
        self._inputs={}
        self._outputs={}
        self.declareInputs()
        self.declareOutputs()

    def input(self, name):
        try:
            return self._inputs[name]
        except:
            pass

    def output(self, name):
        try:
            return self._outputs[name]
        except:
            pass

    def str2value(self, data):
        if data is None:
            return None
        data=data.strip().lower()
        if data in ['1', 'true', 'on', 'yes']:
            return 1
        if data in ['0', 'false', 'off', 'no']:
            return 0
        try:
            return float(data)
        except:
            pass
        return 0

    def getInputValue(self, name, default=None):
        value=self.input(name)
        if value is not None:
            try:
                return value.value
            except:
                return value
        return default

    def onLoadInputs(self, xml: XMLConfig):
        inputs=xml.children('input')
        if inputs:
            for i in inputs:
                name=i.get('name')
                # check if input is available
                if self.input(name) is not None:
                    # 1. check if value is a ref to an existing mbio value
                    value=self.getMBIO().value(i.get('source'))
                    if value is not None:
                        self._inputs[name]=value
                        continue

                    # 2. value is a constant
                    value=self.str2value(i.get('value'))
                    if value is not None:
                        self._inputs[name]=value
                        continue

        # declare mbio values for still unlinked inputs
        for i in self._inputs.values():
            try:
                name=i['name']
                if i['type']=='analog':
                    value=self.value('i_%s' % name, unit=i['unit'], resolution=i['resolution'], default=i['default'], writable=True)
                    self._inputs[name]=value
                    continue
                if i['type']=='digital':
                    value=self.valueDigital('i_%s' % name, default=i['default'], writable=True)
                    self._inputs[name]=value
                    continue
            except:
                pass

    def onLoadOutputs(self, xml: XMLConfig):
        outputs=xml.children('output')
        if outputs:
            for o in outputs:
                name=o.get('name')
                # check if output is available
                if self.output(name) is not None:
                    # TODO:
                    pass

    def configure(self, xml: XMLConfig):
        pass

    def onLoad(self, xml: XMLConfig):
        self.onLoadInputs(xml)
        self.onLoadOutputs(xml)
        self.configure(xml)

    def poweron(self):
        return True

    def poweroff(self):
        return True

    def run(self):
        return 5.0


class MBIOTaskVPLCTest(MBIOTaskVPLC):
    def declareInputs(self):
        self.declareAnalogInput('t', 'C')
        self.declareAnalogInput('spt', 'C')
        self.declareDigitalInput('heat')
        self.declareDigitalInput('cool')

    def declareOutputs(self):
        self.declareAnalogOutput('vheat', '%')
        self.declareAnalogOutput('vcool', '%')

    def configure(self, xml: XMLConfig):
        pass

    def poweron(self):
        return True

    def poweroff(self):
        return True

    def run(self):
        self.values.o_vheat.updateValue(self.getInputValue('spt'))
        return 1.0


if __name__ == "__main__":
    pass



#!/bin/python

from .task import MBIOTask
from .xmlconfig import XMLConfig

import requests
import io
import csv

# https://data.geo.admin.ch/ch.meteoschweiz.messwerte-aktuell/VQHA80.csv
# https://data.geo.admin.ch/ch.meteoschweiz.messwerte-aktuell/info/VQHA80_fr.txt

# tre200s0     °C             température de l'air à 2 m du sol; valeur instantanée
# rre150z0     mm             précipitation: sommation sur 10 minutes
# sre000z0     min            durée d'ensoleillement; sommation sur 10 minutes
# gre000z0     W/m²           rayonnement global; moyenne sur 10 minutes
# ure200s0     %              humidité de l'air relative à 2 m su sol; valeur instantanée
# tde200s0     °C             point de rosée à 2 m du sol; valeur instantanée
# dkl010z0     °              direction du vent; moyenne sur 10 minutes
# fu3010z0     km/h           vitesse du vent; moyenne sur 10 minutes
# fu3010z1     km/h           rafale (intégration 1 s); maximum
# prestas0     hPa            pression atmosphérique à l'altitude de la station (QFE); valeur instantanée
# pp0qffs0     hPa            pression atmosphérique réduite au niveau de la mer (QFF); valeur instantanée
# pp0qnhs0     hPa            pression atmosphérique réduite avec l'atmosphère standard (QNH); valeur instantanée
# ppz850s0     gpm            altitude géopotentielle de la surface de 850 hPa; valeur instantanée
# ppz700s0     gpm            altitude géopotentielle de la surface de 700 hPa; valeur instantanée
# dv1towz0     °              direction du vent vectorielle; instrument 1; moyenne sur 10 minutes
# fu3towz0     km/h           vitesse du vent, tour; moyenne sur 10 minutes
# fu3towz1     km/h           rafale (intégration 1 s), tour; maximum
# ta1tows0     °C             température de l'air instrument 1
# uretows0     %              humidité de l'air relative tour; valeur instantanée
# tdetows0     °C             point de rosée, tour


class MBIOTaskMeteoSuisse(MBIOTask):
    def initName(self):
        return 'weather'

    def onInit(self):
        self._timeoutRefresh=0
        self._retry=3
        self._stations={}

    def onLoad(self, xml: XMLConfig):
        self.valueDigital('comerr', default=False)

        items=xml.children('station')
        if items:
            for item in items:
                name=item.get('name')
                if name and not self._stations.get(name):
                    data={}
                    data['t']=self.value('%s_t' % name, unit='C', resolution=0.1)
                    data['r']=self.value('%s_r' % name, unit='W/m2', resolution=1)
                    data['hr']=self.value('%s_hr' % name, unit='%', resolution=1)
                    data['dewp']=self.value('%s_dewp' % name, unit='C', resolution=0.1)
                    data['wind']=self.value('%s_wind' % name, unit='ms', resolution=0.1)
                    data['rain']=self.value('%s_rain' % name, unit='mm', resolution=0.1)
                    self._stations[name.lower()]=data

    def poweron(self):
        return True

    def poweroff(self):
        return True

    def url(self):
        return 'https://data.geo.admin.ch/ch.meteoschweiz.messwerte-aktuell/VQHA80.csv'

    def decodeValueFloat(self, value, data, datakey, factor=1.0):
        if value is not None and data:
            try:
                v=data[datakey]
                if v and v!='-':
                    v=float(v)*factor
                    value.updateValue(v)
                    value.setError(False)
                    return True
            except:
                pass
            value.setError(True)
            return False

    def run(self):
        if self.isTimeout(self._timeoutRefresh):
            self._timeoutRefresh=self.timeout(60)
            error=False

            try:
                url=self.url()
                self.logger.debug('%s(%s)' % (self.__class__.__name__, url))
                r=requests.get(self.url(), timeout=5)
                if r and r.ok:
                    # self.logger.debug(r.text)
                    f=io.StringIO(r.text)
                    reader=csv.reader(f, delimiter=';')
                    headers=[]
                    for row in reader:
                        if not headers:
                            headers.extend(row)
                            continue

                        station=self._stations.get(row[0].lower())
                        if station:
                            data={}
                            for n in range(2, len(headers)):
                                data[headers[n].lower()]=row[n]

                            if not self.decodeValueFloat(station['t'], data, 'tre200s0'):
                                error=True

                            if not self.decodeValueFloat(station['r'], data, 'gre000z0'):
                                error=True

                            if not self.decodeValueFloat(station['hr'], data, 'ure200s0'):
                                error=True

                            if not self.decodeValueFloat(station['dewp'], data, 'tde200s0'):
                                error=True

                            if not self.decodeValueFloat(station['wind'], data, 'fu3010z0', 1.0/3.6):
                                error=True

                            if not self.decodeValueFloat(station['rain'], data, 'rre150z0'):
                                error=True
                else:
                    error=True
                    self.logger.error('%s(%s)' % (self.__class__.__name__, url))

            except:
                # self.logger.exception('meteo')
                error=True

            if not error:
                self._timeoutRefresh=self.timeout(60*10)
                self._retry=3
                self.values.comerr.updateValue(False)
            else:
                self._timeoutRefresh=self.timeout(60)
                if self._retry>0:
                    self._retry-=1
                    if self._retry==0:
                        self.values.comerr.updateValue(True)
                        for station in self._stations.keys():
                            for value in self._stations[station].values():
                                value.setError(True)

        return 5.0


if __name__ == "__main__":
    pass
